/*
 * lib/img.c, part of Wlib, part of W
 * (C) 94-07/96 by Torsten Scherer (TeSche)
 * itschere@techfak.uni-bielefeld.de
 *
 * routines to read a monochrome IMG file as a BITMAP
 *
 * - fixed IMG decoding (IMGs are byte aligned, BITMAPs now long aligned).
 *   8/96 ++eero
 * - Added byte conversions for header values.
 * - changed error messages to trace stuff.
 *   5/97 ++eero
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>		/* byte conversion stuff */
#include <sys/stat.h>
#include "Wlib.h"
#include "proto.h"

/*
 * IMG header definition
 */

typedef struct {
  ushort ver;		/* version (1) */
  ushort hdrlen;	/* header lenght as shorts (8) */
  ushort planes;	/* number of bitplanes     (1) */
  ushort patlen;	/* lenght of patterns      (2) */
  ushort pixwidth;	/* pixel width in um       (372) */
  ushort pixheight;	/* pixel height in um      (372) */
  ushort scanwidth;	/* pixels / line (= (w+7)/8 bytes) */
  ushort scanlines;	/* total number of lines */
} IMGHDR;


/*
 * routines for loading a monochrome IMG picture
 */

static ushort *load(char *fname, char **error)
{
  struct stat st;
  ushort *ptr;
  short fh;

  if (stat(fname, &st)) {
    *error = "can't stat %s\n", fname;
    return NULL;
  }

  if (st.st_size < sizeof(IMGHDR)) {
    *error = "file too short";
    return NULL;
  }

  if (!(ptr = malloc(st.st_size))) {
    *error = "file alloc failed";
    return NULL;
  }

  if ((fh = open(fname, O_RDONLY, 0)) < 0) {
    *error = "file open failed";
    free(ptr);
    return NULL;
  }

  if (read(fh, ptr, st.st_size) != st.st_size) {
    *error = "file load failed";
    close(fh);
    free(ptr);
    return NULL;
  }

  close(fh);
  return ptr;
}


static long decode(uchar *data, short scanbytes, short patlen,
                   short realbytes, short scanlines, uchar *pic)
{
  short cnt, i, j, bytecnt;
  uchar first, *repptr = data, repcnt = 0;

  realbytes -= scanbytes;

  while (scanlines--) {

    if (repcnt) {
      repcnt--;
      data = repptr;
    }

    /* is there a repetition header? */

    if (!*data && !*(data+1) && (*(data+2) == 0xff)) {
      data += 3;
      repcnt = *data++;
      repptr = data;
    }

    bytecnt = scanbytes;
    while (bytecnt > 0)
      switch (first = *data++) {

      case 0: /* pattern run */
        cnt = *data++;
	for (i=0; i<cnt; i++)
	  for (j=0; j<patlen; j++)
	    *pic++ = *(data+j);
	data += patlen;
	bytecnt -= (cnt * patlen);
	break;

      case 128: /* bit string */
	cnt = *data++;
	for (i=0; i<cnt; i++)
	  *pic++ = *data++;
	bytecnt -= cnt;
	break;

      default: /* solid run */
	cnt = first & 127;
	if (first & 128)
	  for (i=0; i<cnt; i++)
	    *pic++ = 255;
	else
	  for (i=0; i<cnt; i++)
	    *pic++ = 0;
	bytecnt -= cnt;
    }
    if(bytecnt)
      return 1;		/* error: line encoded wrong */

    pic += realbytes;
  }

  return 0;
}


/*
 * what we're going to export...
 */

BITMAP *w_readimg(char *fname, short *width, short *height)
{
  ushort *file, *data;
  uchar *picture;
  IMGHDR *imghdr;
  BITMAP *ret;
  char *error;
  int idx;

  TRACESTART();

  if (!(file = load(fname, &error))) {
    TRACEPRINT(("w_readimg(%s,%p,%p) -> %s\n", fname, width, height, error));
    TRACEEND();
    return NULL;
  }

  data = file;
  for (idx = 0; idx < 8; idx++) {
	 *data = ntohs(*data);
	 data++;
  }
  imghdr = (IMGHDR *)file;

  if (imghdr->ver != 1 || imghdr->hdrlen < 8 || imghdr->planes != 1) {
    free(file);
    TRACEPRINT(("w_readimg(%s,%p,%p) -> not a monochrome IMG\n", fname, width, height));
    TRACEEND();
    return NULL;
  }

  /* enough memory to load picture? */
  ret = w_allocbm(imghdr->scanwidth, imghdr->scanlines, BM_PACKEDMONO, 2);
  if (!ret) {
    free(file);
    TRACEPRINT(("w_readimg(%s,%p,%p) -> bm alloc failed\n", fname, width, height));
    TRACEEND();
    return NULL;
  }

  picture = (uchar *)(file + imghdr->hdrlen);
  if (decode(picture, (imghdr->scanwidth + 7) >> 3, imghdr->patlen,
	     ret->unitsize * ret->upl, ret->height, ret->data) < 0) {
    w_freebm(ret);
    free(file);
    TRACEPRINT(("w_readimg(%s,%p,%p) -> IMG decode failed\n", fname, width, height));
    TRACEEND();
    return NULL;
  }

  free(file);
  *width = ret->width;
  *height = ret->height;

  TRACEPRINT(("w_readimg(%s,%p,%p) -> %p\n", fname, width, height, ret));
  TRACEEND();
  return ret;
}

