/*
 * this file is part of "The W Toolkit".
 *
 * (W) 1996, Kay Roemer.
 *
 * VT 52 widget with scrollback buffer and xterm-like selection.
 *
 * TODO:
 * - convert to vt100/ansi
 *
 * $Id: vt.c,v 1.3 1996/12/07 19:27:52 yak Exp $
 */
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#include <assert.h>
#include <Wlib.h>
#include "Wt.h"
#include "config.h"	/* server MAXFAMILYSIZE */
#include "toolkit.h"


#define ATTR_SL		1 /* slanted */
#define ATTR_BD		2 /* bold */
#define ATTR_RV		4 /* reverse */
#define ATTR_UL		8 /* underlined */
#define ATTR_FONT_MASK	(ATTR_SL|ATTR_BD)
#define ATTR_STYLE_MASK	(ATTR_RV|ATTR_UL)

typedef struct {
  u_char attr;
  u_char c;
} vtchar_t;

#define BUFSIZE		80

typedef struct _vt_widget_t {
  widget_t w;
  short    is_open;
  short    is_realized;

  long	   clicktime;	/* time of last mouse click */
  short	   nclicks;	/* # mouse clicks since clicktime */

  long	   selx0,sely0; /* selection start */
  long	   selx1,sely1; /* selection end */
  long	   seltimer;	/* selection timer */
  short	   inselect;	/* selection valid */

  long	   histsize;	/* max. history size */
  long	   curhistsize;	/* lines currently in history */

  long     wd, ht;	/* visible width and height in chars */
  long	   screenht;	/* ht + histsize */
  long     screen_size; /* wd * screenht */
  vtchar_t *screen;	/* circular screen buffer */
  long     offs;	/* screen[offs*wd] is top left char in vt window */
  long	   curoffs;	/* screen[curoffs*wd] is top left char currently
			   shown in the window (!= offs only if user has
			   scrolled back the window */

  long     curx, cury;	/* cursor pos */
  long     savx, savy;	/* saved cursor pos */
  u_char   curvis;	/* cursor currently shown */
  u_char   curon;	/* cursor on/off */
  u_char   curblink;	/* cursor blinking on/off */
  u_char   autowrap;	/* autowrap on/off */
  u_char   visbell;	/* visual bell on/off */
  u_char   curattr;	/* display attributes */
  WFONT	   *fonts[ATTR_FONT_MASK+1];
  char     fontfamily[MAXFAMILYNAME];
  short	   fontht;	/* font infos */
  short    fontwd;

  long     outx, outy;
  u_char   outbuf[BUFSIZE+1];
  u_char   *outbufp;	/* output buffering */

  void     (*outfn) (struct _vt_widget_t *, u_char c);

  void	   (*hist_cb) (widget_t *w, long histpos, long histsize,
  		       long rows, long cols, wt_opaque_t *);
  short usrwd, usrht, usrcols, usrrows;
} vt_widget_t;


static void putc_normal (vt_widget_t *, u_char c);
static void putc_esc    (vt_widget_t *, u_char c);
static void putc_esc_y1 (vt_widget_t *, u_char c);
static void putc_esc_y2 (vt_widget_t *, u_char c);
static void putc_esc_c  (vt_widget_t *, u_char c);
static void putc_esc_b  (vt_widget_t *, u_char c);


/**********************************************************************/

static inline vtchar_t *
vt_screen_pos (vt_widget_t *w, int x, int y)
{
  if ((y += w->curoffs) >= w->screenht)
    y -= w->screenht;
  return &w->screen[y * w->wd + x];
}

static inline vtchar_t *
vt_screen_nl (vt_widget_t *w, vtchar_t *vcp)
{
  if (vcp - w->screen >= w->screen_size)
    vcp -= w->screen_size;
  return vcp;
}

static inline vtchar_t *
vt_screen_pl (vt_widget_t *w, vtchar_t *vcp)
{
  if (vcp < w->screen)
    vcp += w->screen_size;
  return vcp;
}

static int
vt_loadfonts (vt_widget_t *w, char *basename, short fsize)
{
  WFONT *fp;

  w->fontht=fsize;
  if (!(fp = wt_loadfont (basename, fsize, 0, 1))) {
      printf("vt_loadfont of %s size %d failed!\n",basename,fsize);
      return -1;
  }
  w->fonts[0] = fp;

  if (!(w->fonts[ATTR_BD] = wt_loadfont (basename, w->fontht, F_BOLD, 1)))
    w->fonts[ATTR_BD] = wt_loadfont (basename, w->fontht, 0, 1);

  if (!(w->fonts[ATTR_SL] = wt_loadfont (basename, w->fontht, F_ITALIC, 1)))
    w->fonts[ATTR_SL] = wt_loadfont (basename, w->fontht, 0, 1);

  if (!(w->fonts[ATTR_SL|ATTR_BD] = wt_loadfont (basename, w->fontht, F_BOLD | F_ITALIC, 1)))
    w->fonts[ATTR_SL|ATTR_BD] = wt_loadfont (basename, w->fontht, 0, 1);

  strncpy (w->fontfamily, w->fonts[0]->family, sizeof (w->fontfamily));
  w->fontfamily[sizeof (w->fontfamily) - 1] = '\0';

  w->fontwd = w->fonts[0]->maxwidth;
  w->fontht = w->fonts[0]->height;
  fprintf(stderr,"fontfamily='%s'\n",w->fonts[0]->family);
  fprintf(stderr,"height='%d'\n",w->fonts[0]->height);
  fprintf(stderr,"width='%d'\n",w->fonts[0]->maxwidth);
  return 0;
}

static void
vt_unloadfonts (vt_widget_t *w)
{
  int i;

  for (i = 0; i <= ATTR_FONT_MASK; ++i) {
    if (w->fonts[i]) {
      w_unloadfont (w->fonts[i]);
    }
  }
}

static int
vt_initialize (vt_widget_t *w, long wd, long ht, long screenht)
{
  vtchar_t *vcp;
  int i;

  w->wd = wd;
  w->ht = ht;
  w->screenht = screenht;
  w->offs = 0;
  w->screen_size = wd * screenht;
  w->screen = malloc (w->screen_size * sizeof (vtchar_t));
  if (!w->screen)
    return -1;

  for (i = w->screen_size, vcp = w->screen; --i >= 0; ++vcp) {
    vcp->attr = 0;
    vcp->c = ' ';
  }
  w->curx = w->savx = 0;
  w->cury = w->savy = 0;

  w->curvis = 0;
  w->curon = 1;
  w->curblink = 0;
  w->autowrap = 1;
  w->curattr = 0;

  w->outx = w->outy = 0;
  w->outbufp = w->outbuf;

  w->outfn = putc_normal;
  return 0;
}

static void
vt_exit (vt_widget_t *w)
{
  vt_unloadfonts (w);
  if (w->screen)
    free (w->screen);
}

static void
vt_flush (vt_widget_t *w)
{
  vtchar_t *vcp;
  char *cp;

  if (w->outbufp > w->outbuf) {
    *w->outbufp = 0;
    w_printstring (w->w.win, w->outx*w->fontwd, w->outy*w->fontht, w->outbuf);

    vcp = vt_screen_pos (w, w->outx, w->outy);
    for (cp = w->outbuf; *cp; ++vcp) {
      vcp->attr = w->curattr;
      vcp->c = *cp++;
    }
    w->outbufp = w->outbuf;
  }
}

static inline void
vt_addchar (vt_widget_t *w, u_char c)
{
  if (w->outbufp - w->outbuf >= BUFSIZE)
    vt_flush (w);
  if (w->outbuf == w->outbufp) {
    w->outx = w->curx;
    w->outy = w->cury;
  }
  *w->outbufp++ = c;
}

static void
vt_cursor_show (vt_widget_t *w)
{
  if (w->curon && !w->curvis && w->offs == w->curoffs && !w->inselect) {
    w->curvis = 1;
    w_setmode (w->w.win, M_INVERS);
    w_pbox (w->w.win, w->curx * w->fontwd, w->cury * w->fontht,
    	    w->fontwd, w->fontht);
  }
}

static void
vt_cursor_hide (vt_widget_t *w)
{
  if (w->curvis) {
    w->curvis = 0;
    w_setmode (w->w.win, M_INVERS);
    w_pbox (w->w.win, w->curx * w->fontwd, w->cury * w->fontht,
    	    w->fontwd, w->fontht);
  }
}

static void
vt_setattr (vt_widget_t *w, u_char attr)
{
  if ((attr ^ w->curattr) & ATTR_FONT_MASK) {
    w_setfont (w->w.win, w->fonts[attr & ATTR_FONT_MASK]);
  }
  if ((attr ^ w->curattr) & ATTR_STYLE_MASK) {
    short style = 0;
    if (attr & ATTR_UL)
      style |= F_UNDERLINE;
    if (attr & ATTR_RV)
      style |= F_REVERSE;
    w_settextstyle (w->w.win, style);
  }
  w->curattr = attr;
}

#define SWAP(x, y) { (x) ^= (y); (y) ^= (x); (x) ^= (y); }

static void
vt_select_paint (vt_widget_t *w, long x0, long y0, long x1, long y1)
{
  long ht, wd;

  w_setmode (w->w.win, M_INVERS);

  if (y0 > y1) {
    SWAP(x0, x1);
    SWAP(y0, y1);
  } else if (y0 == y1 && x0 > x1) {
    SWAP(x0, x1);
  }

  ht = w->fontht;
  wd = w->fontwd;

  if (y0 == y1) {
    w_pbox (w->w.win, x0*wd, y0*ht, (x1-x0)*wd, ht);
  } else {
    if (x0 < w->wd)
      w_pbox (w->w.win, x0*wd, y0*ht, (w->wd-x0)*wd, ht);
    if (y0+1 < y1)
      w_pbox (w->w.win, 0, (y0+1)*ht, w->wd*wd, (y1-y0-1)*ht);
    w_pbox (w->w.win, 0, y1*ht, x1*wd, ht);
  }
}

static void
vt_unselect (vt_widget_t *w)
{
  if (w->inselect) {
    vt_select_paint (w, w->selx0, w->sely0, w->selx1, w->sely1);
    if (w->seltimer >= 0) {
      wt_deltimeout (w->seltimer);
      w->seltimer = -1;
    }
    w->inselect = 0;
  }
  vt_cursor_show (w);
}

static void
vt_select_start (vt_widget_t *w, long x, long y)
{
  vt_unselect (w);
  vt_cursor_hide (w);
  w->inselect = 1;

  if ((x /= w->fontwd) < 0)
    x = 0;
  else if (x > w->wd)
    x = w->wd;
  if ((y /= w->fontht) < 0)
    y = 0;
  else if (y >= w->ht)
    y = w->ht-1;

  w->selx0 = w->selx1 = x;
  w->sely0 = w->sely1 = y;
  vt_select_paint (w, x, y, x, y);
}

static void
vt_select_drag (vt_widget_t *w, long x, long y)
{
  if (!w->inselect)
    return;

  if ((x /= w->fontwd) < 0)
    x = 0;
  else if (x > w->wd)
    x = w->wd;
  if ((y /= w->fontht) < 0)
    y = 0;
  else if (y >= w->ht)
    y = w->ht-1;

  if (x != w->selx1 || y != w->sely1) {
    vt_select_paint (w, w->selx1, w->sely1, x, y);
    w->selx1 = x;
    w->sely1 = y;
  }
}

static void
vt_select_end (vt_widget_t *w)
{
  long x0, y0, x1, y1, len, i, j;
  char *buf, *cp;
  vtchar_t *vcp;

  if (!w->inselect)
    return;

  if (w->seltimer >= 0) {
    wt_deltimeout (w->seltimer);
    w->seltimer = -1;
  }

  x0 = w->selx0;
  y0 = w->sely0;
  x1 = w->selx1;
  y1 = w->sely1;

  if (y0 > y1) {
    SWAP(x0, x1);
    SWAP(y0, y1);
  } else if (y0 == y1 && x0 > x1) {
    SWAP(x0, x1);
  }

  len = (y1 - y0 + 1) * (w->wd + 1);
  cp = buf = malloc (len+1);
  assert (buf);

  for (i = y0; i <= y1; ++i) {
    long first = (i == y0) ? x0 : 0;
    long last  = (i == y1) ? x1 : w->wd;
    vcp = vt_screen_pos (w, 0, i);

    while (last > first && vcp[last-1].c == ' ')
      --last;
    for (j = first; j < last; ++j)
      *cp++ = vcp[j].c;
    if (i < y1 || x1 == w->wd)
      *cp++ = '\n';
  }
  *cp = 0;

  if (w_putselection (W_SEL_TEXT, buf, strlen (buf)) < 0)
    w_beep ();
  free (buf);

  if (y0 == y1 && x0 == x1)
    vt_unselect (w);
}

static void
vt_select_word (vt_widget_t *w, long x, long y)
{
  vtchar_t *vcp;
  long i, isws;

  vt_unselect (w);
  vt_cursor_hide (w);
  w->inselect = 1;

  if ((x /= w->fontwd) < 0)
    x = 0;
  else if (x >= w->wd)
    x = w->wd-1;
  if ((y /= w->fontht) < 0)
    y = 0;
  else if (y >= w->ht)
    y = w->ht-1;

  vcp = vt_screen_pos (w, x, y);
  isws = !!isspace (vcp->c);
  for (i = x; i >= 0; --i, --vcp) {
    if (isws != !!isspace(vcp->c))
      break;
  }
  w->selx0 = i+1;
  w->sely0 = y;

  vcp = vt_screen_pos (w, x, y);
  for (i = x; i < w->wd; ++i, ++vcp) {
    if (isws != !!isspace(vcp->c))
      break;
  }
  w->selx1 = i;
  w->sely1 = y;

  vt_select_paint (w, w->selx0, w->sely0, w->selx1, w->sely1);
  vt_select_end (w);
}

static void
vt_select_line (vt_widget_t *w, long x, long y)
{
  vt_unselect (w);
  vt_cursor_hide (w);
  w->inselect = 1;

  if ((y /= w->fontht) < 0)
    y = 0;
  else if (y >= w->ht)
    y = w->ht-1;

  w->selx0 = 0;
  w->sely0 = y;
  w->selx1 = w->wd;
  w->sely1 = y;

  vt_select_paint (w, w->selx0, w->sely0, w->selx1, w->sely1);
  vt_select_end (w);
}

static void
vt_redraw (vt_widget_t *w, long ymin, long ymax)
{
  char buf[BUFSIZE+1];
  char *cp;
  u_char attr, saveattr;
  int i, j, xpos, ypos;
  vtchar_t *vcp;

  saveattr = w->curattr;
  cp = buf;
  xpos = 0;
  ypos = ymin * w->fontht;
  vcp = vt_screen_pos (w, 0, ymin);
  attr = vcp->attr;
  vt_setattr (w, attr);

  for (i = ymin; i < ymax; ++i) {
    for (j = 0; j < w->wd; ++j, ++vcp) {
      if (attr != vcp->attr || cp - buf >= BUFSIZE) {
	*cp = 0;
	w_printstring (w->w.win, xpos, ypos, buf);
	xpos += (cp - buf) * w->fontwd;
	cp = buf;
	attr = vcp->attr;
	vt_setattr (w, attr);
      }
      *cp++ = vcp->c;
    }
    *cp = 0;
    cp = buf;
    w_printstring (w->w.win, xpos, ypos, buf);
    xpos = 0;
    ypos += w->fontht;
    vcp = vt_screen_nl (w, vcp);
  }
  vt_setattr (w, saveattr);
}

static void
vt_vmove (vt_widget_t *w, long offs)
{
  long oldoffs, dist, ymin, ymax;

  /*
   * make sure you cannot scroll back more than size of history
   */
  if (offs < 0)
    offs = 0;
  else if (offs > w->curhistsize)
    offs = w->curhistsize;
  oldoffs = w->offs - w->curoffs;
  if (oldoffs < 0)
    oldoffs += w->screenht;

  if ((dist = abs (offs - oldoffs)) == 0)
    return;

  vt_unselect (w);
  vt_cursor_hide (w);

  if (dist < w->ht) {
    if (offs > oldoffs) {
      /* move backwards in history */
      ymin = 0;
      ymax = dist;
      dist *= w->fontht;
      w_vscroll (w->w.win, 0, 0, w->w.w, w->w.h - dist, dist);
    } else {
      /* move forwards in history */
      ymin = w->ht - dist;
      ymax = w->ht;
      dist *= w->fontht;
      w_vscroll (w->w.win, 0, dist, w->w.w, w->w.h - dist, 0);
    }
  } else {
    ymin = 0;
    ymax = w->ht;
  }

  if ((w->curoffs = w->offs - offs) < 0)
    w->curoffs += w->screenht;

  vt_redraw (w, ymin, ymax);
  vt_cursor_show (w);

  if (w->hist_cb)
    (*w->hist_cb) ((widget_t *)w, offs, -1, -1, -1, NULL);
}

static void
vt_output (vt_widget_t *w, u_char *cp, long len)
{
  long ohistsize = w->curhistsize;

  vt_vmove (w, 0);
  vt_unselect (w);

  vt_cursor_hide (w);
  while (--len >= 0)
    (*w->outfn) (w, *cp++);
  vt_flush (w);
  vt_cursor_show (w);

  if (ohistsize != w->curhistsize && w->hist_cb)
    (*w->hist_cb) ((widget_t *)w, -1, w->curhistsize, -1, -1, NULL);
}

static int
vt_resize (vt_widget_t *w, long wd, long ht, long screenht)
{
  vtchar_t *nscreen, *vcp1, *vcp2;
  int i, j;

  if (wd < 1)
    wd = 1;
  if (ht < 1)
    ht = 1;
  if (screenht < ht)
    screenht = ht;

  if (wd == w->wd && ht == w->ht && screenht == w->screenht)
    return 0;

  nscreen = malloc (wd * screenht * sizeof (vtchar_t));
  if (!nscreen)
    return -1;

  vt_flush (w);
  vt_vmove (w, 0);
  vt_unselect (w);
  vt_cursor_hide (w);

  w_resize (w->w.win, wd * w->fontwd, ht * w->fontht);

  for (i = wd * screenht; --i >= 0; ) {
    nscreen[i].attr = 0;
    nscreen[i].c = ' ';
  }
  for (i = MIN (w->screenht, screenht); --i >= 0; ) {
    j = (MIN (w->ht, ht) - 1 - i) % screenht;
    if (j < 0)
      j += screenht;
    vcp1 = &nscreen[j * wd];

    j = (w->curoffs + MIN (w->ht, ht) - 1 - i) % w->screenht;
    if (j < 0)
      j += w->screenht;
    vcp2 = &w->screen[j * w->wd];

    for (j = MIN (w->wd, wd); --j >= 0; ) {
      *vcp1++ = *vcp2++;
    }
  }
  w->wd = wd;
  w->ht = ht;
  w->screenht = screenht;
  w->screen_size = wd * screenht;
  w->curoffs = w->offs = 0;
  free (w->screen);
  w->screen = nscreen;

  w->curx = MIN (w->wd-1, w->curx);
  w->cury = MIN (w->ht-1, w->cury);
  w->savx = MIN (w->wd-1, w->savx);
  w->savy = MIN (w->ht-1, w->savy);

  vt_cursor_show (w);

  if (w->hist_cb)
    (*w->hist_cb) ((widget_t *)w, -1, w->curhistsize, w->wd, w->ht, NULL);

  return 0;
}

/************************************************************************/

static void
vt_bell (vt_widget_t *w)
{
  if (w->visbell) {
    w_setmode (w->w.win, M_INVERS);
    w_pbox (w->w.win, 0, 0, w->w.w, w->w.h);
    w_test (w->w.win, 0, 0);
    w_pbox (w->w.win, 0, 0, w->w.w, w->w.h);
  } else {
    w_beep ();
  }
}

static inline void
vt_cursor_on (vt_widget_t *w)
{
  w->curon = 1;
}

static inline void
vt_cursor_off (vt_widget_t *w)
{
  w->curon = 0;
}

static inline void
vt_cursor_xpos (vt_widget_t *w, long x)
{
  vt_flush (w);
  w->curx = (x - 32) & 0x7f;
  if (w->curx >= w->wd)
    w->curx = w->wd-1;
  else if (w->curx < 0)
    w->curx = 0;
}

static inline void
vt_cursor_ypos (vt_widget_t *w, long y)
{
  vt_flush (w);
  w->cury = (y - 32) & 0x7f;
  if (w->cury >= w->ht)
    w->cury = w->ht-1;
  else if (w->cury < 0)
    w->cury = 0;
}

static inline void
vt_cursor_up (vt_widget_t *w)
{
  vt_flush (w);
  if (--w->cury < 0)
    w->cury = 0;
}

static inline void
vt_cursor_down (vt_widget_t *w)
{
  vt_flush (w);
  if (++w->cury >= w->ht)
    w->cury = w->ht-1;
}

static inline void
vt_cursor_left (vt_widget_t *w)
{
  vt_flush (w);
  if (--w->curx < 0)
    w->curx = 0;
}

static inline void
vt_cursor_right (vt_widget_t *w)
{
  vt_flush (w);
  if (++w->curx >= w->wd)
    w->curx = w->wd-1;
}

static inline void
vt_cursor_home (vt_widget_t *w)
{
  vt_flush (w);
  w->curx = w->cury = 0;
}

static inline void
vt_cursor_save (vt_widget_t *w)
{
  w->savx = w->curx;
  w->savy = w->cury;
}

static inline void
vt_cursor_restore (vt_widget_t *w)
{
  vt_flush (w);
  w->curx = w->savx;
  w->cury = w->savy;
}

static inline void
vt_clear_screen (vt_widget_t *w)
{
  int i, j;
  vtchar_t *vcp;

  vt_flush (w);
  vcp = vt_screen_pos (w, 0, 0);
  for (i = 0; i < w->ht; ++i) {
    for (j = 0; j < w->wd; ++j, ++vcp) {
      vcp->attr = 0;
      vcp->c = ' ';
    }
    vcp = vt_screen_nl (w, vcp);
  }
  w_setmode (w->w.win, M_CLEAR);
  w_pbox (w->w.win, 0, 0, w->w.w, w->w.h);
}

static inline void
vt_clear_line (vt_widget_t *w)
{
  int i;
  vtchar_t *vcp;

  vt_flush (w);
  vcp = vt_screen_pos (w, 0, w->cury);
  for (i = 0; i < w->wd; ++i, ++vcp) {
    vcp->attr = 0;
    vcp->c = ' ';
  }
  w_setmode (w->w.win, M_CLEAR);
  w_pbox (w->w.win, 0, w->cury * w->fontht, w->w.w, w->fontht);
}

static inline void
vt_clear_eop (vt_widget_t *w)
{
  int i, j;
  vtchar_t *vcp;

  vt_flush (w);
  vcp = vt_screen_pos (w, w->curx, w->cury);
  for (i = w->curx; i < w->wd; ++i, ++vcp) {
    vcp->attr = 0;
    vcp->c = ' ';
  }
  vcp = vt_screen_nl (w, vcp);
  for (i = w->cury+1; i < w->ht; ++i) {
    for (j = 0; j < w->wd; ++j, ++vcp) {
      vcp->attr = 0;
      vcp->c = ' ';
    }
    vcp = vt_screen_nl (w, vcp);
  }
  w_setmode (w->w.win, M_CLEAR);
  if (w->cury < w->ht-1) {
    w_pbox (w->w.win, 0, (w->cury + 1) * w->fontht,
	    w->w.w, (w->ht - 1 - w->cury) * w->fontht);
  }
  if (w->curx < w->wd-1) {
    w_pbox (w->w.win, w->curx * w->fontwd, w->cury * w->fontht,
  	    (w->wd - w->curx) * w->fontwd, w->fontht);
  }
}

static inline void
vt_clear_sop (vt_widget_t *w)
{
  int i, j;
  vtchar_t *vcp;

  vt_flush (w);
  vcp = vt_screen_pos (w, 0, 0);
  for (i = 0; i < w->cury; ++i) {
    for (j = 0; j < w->wd; ++j, ++vcp) {
      vcp->attr = 0;
      vcp->c = ' ';
    }
    vcp = vt_screen_nl (w, vcp);
  }
  for (i = 0; i < w->curx; ++i, ++vcp) {
    vcp->attr = 0;
    vcp->c = ' ';
  }
  w_setmode (w->w.win, M_CLEAR);
  if (w->cury > 0)
    w_pbox (w->w.win, 0, 0, w->w.w, w->cury * w->fontht);
  if (w->curx > 0)
    w_pbox (w->w.win, 0, w->cury * w->fontht, w->curx * w->fontwd, w->fontht);
}

static inline void
vt_clear_eol (vt_widget_t *w)
{
  int i;
  vtchar_t *vcp;

  vt_flush (w);
  vcp = vt_screen_pos (w, w->curx, w->cury);
  for (i = w->curx; i < w->wd; ++i, ++vcp) {
    vcp->attr = 0;
    vcp->c = ' ';
  }
  if (w->curx < w->wd-1) {
    w_setmode (w->w.win, M_CLEAR);
    w_pbox (w->w.win, w->curx * w->fontwd, w->cury * w->fontht,
  	    (w->wd - w->curx) * w->fontwd, w->fontht);
  }
}

static inline void
vt_clear_sol (vt_widget_t *w)
{
  int i;
  vtchar_t *vcp;

  vt_flush (w);
  vcp = vt_screen_pos (w, 0, w->cury);
  for (i = 0; i < w->curx; ++i, ++vcp) {
    vcp->attr = 0;
    vcp->c = ' ';
  }
  if (w->curx > 0) {
    w_setmode (w->w.win, M_CLEAR);
    w_pbox (w->w.win, 0, w->cury * w->fontht, w->curx * w->fontwd, w->fontht);
  }
}

static inline void
vt_insert_line (vt_widget_t *w)
{
  int i, j;
  vtchar_t *vcp1, *vcp2;

  vt_flush (w);
  if (w->cury < w->ht-1) {
    vcp1 = vt_screen_pos (w, w->wd-1, w->ht-2);
    vcp2 = vt_screen_pos (w, w->wd-1, w->ht-1);

    for (i = w->cury; i < w->ht-1; ++i) {
      for (j = 0; j < w->wd; ++j)
        *vcp2-- = *vcp1--;
      vcp1 = vt_screen_pl (w, vcp1);
      vcp2 = vt_screen_pl (w, vcp2);
    }
    w_vscroll (w->w.win, 0, w->cury * w->fontht,
	       w->w.w, (w->ht - 1 - w->cury) * w->fontht,
	       (w->cury + 1) * w->fontht);
  }
  vcp1 = vt_screen_pos (w, 0, w->cury);
  for (i = 0; i < w->wd; ++i, ++vcp1) {
    vcp1->attr = 0;
    vcp1->c = ' ';
  }
  w_setmode (w->w.win, M_CLEAR);
  w_pbox (w->w.win, 0, w->cury * w->fontht, w->w.w, w->fontht);
  w->curx = 0;
}

static inline void
vt_rev_index (vt_widget_t *w)
{
  int i, j;
  vtchar_t *vcp1, *vcp2;

  vt_flush (w);
  if (--w->cury < 0) {
    w->cury = 0;
    vcp1 = vt_screen_pos (w, w->wd-1, w->ht-2);
    vcp2 = vt_screen_pos (w, w->wd-1, w->ht-1);

    for (i = w->cury; i < w->ht-1; ++i) {
      for (j = 0; j < w->wd; ++j)
        *vcp2-- = *vcp1--;
      vcp1 = vt_screen_pl (w, vcp1);
      vcp2 = vt_screen_pl (w, vcp2);
    }
    w_vscroll (w->w.win, 0, w->cury * w->fontht,
	       w->w.w, (w->ht - 1 - w->cury) * w->fontht,
	       (w->cury + 1) * w->fontht);

    vcp1 = vt_screen_pos (w, 0, w->cury);
    for (i = 0; i < w->wd; ++i, ++vcp1) {
      vcp1->attr = 0;
      vcp1->c = ' ';
    }
    w_setmode (w->w.win, M_CLEAR);
    w_pbox (w->w.win, 0, w->cury * w->fontht, w->w.w, w->fontht);
  }
}

static inline void
vt_delete_line (vt_widget_t *w)
{
  int i, j;
  vtchar_t *vcp1, *vcp2;

  vt_flush (w);
  if (w->cury < w->ht-1) {
    vcp1 = vt_screen_pos (w, 0, w->cury+1);
    vcp2 = vt_screen_pos (w, 0, w->cury);

    for (i = w->cury+1; i < w->ht; ++i) {
      for (j = 0; j < w->wd; ++j)
        *vcp2++ = *vcp1++;
      vcp1 = vt_screen_nl (w, vcp1);
      vcp2 = vt_screen_nl (w, vcp2);
    }
    w_vscroll (w->w.win, 0, (w->cury + 1) * w->fontht,
	       w->w.w, (w->ht - 1 - w->cury) * w->fontht,
	       w->cury * w->fontht);
  }
  vcp1 = vt_screen_pos (w, 0, w->ht-1);
  for (i = 0; i < w->wd; ++i, ++vcp1) {
    vcp1->attr = 0;
    vcp1->c = ' ';
  }
  w_setmode (w->w.win, M_CLEAR);
  w_pbox (w->w.win, 0, (w->ht - 1) * w->fontht, w->w.w, w->fontht);
  w->curx = 0;
}

static inline void
vt_scroll (vt_widget_t *w)
{
  int i;
  vtchar_t *vcp;

  vt_flush (w);
  if (++w->offs >= w->screenht)
    w->offs -= w->screenht;
  w->curoffs = w->offs;

  if (w->curhistsize < w->histsize)
    ++w->curhistsize;

  vcp = vt_screen_pos (w, 0, w->ht-1);
  for (i = 0; i < w->wd; ++i, ++vcp) {
    vcp->attr = 0;
    vcp->c = ' ';
  }
  w_vscroll (w->w.win, 0, w->fontht, w->w.w, w->w.h - w->fontht, 0);
  w_setmode (w->w.win, M_CLEAR);
  w_pbox (w->w.win, 0, (w->ht - 1) * w->fontht, w->w.w, w->fontht);
}

static inline void
vt_set_fgcol (vt_widget_t *w, u_char col)
{
  /* not implemented */
  vt_flush (w);
}

static inline void
vt_set_bgcol (vt_widget_t *w, u_char col)
{
  /* not implemented */
  vt_flush (w);
}

static inline void
vt_reverse_on (vt_widget_t *w)
{
  vt_flush (w);
  vt_setattr (w, w->curattr | ATTR_RV);
}

static inline void
vt_reverse_off (vt_widget_t *w)
{
  vt_flush (w);
  vt_setattr (w, w->curattr & ~ATTR_RV);
}

static inline void
vt_bold_on (vt_widget_t *w)
{
  vt_flush (w);
  vt_setattr (w, w->curattr | ATTR_BD);
}

static inline void
vt_bold_off (vt_widget_t *w)
{
  vt_flush (w);
  vt_setattr (w, w->curattr & ~ATTR_BD);
}

static inline void
vt_underline_on (vt_widget_t *w)
{
  vt_flush (w);
  vt_setattr (w, w->curattr | ATTR_UL);
}

static inline void
vt_underline_off (vt_widget_t *w)
{
  vt_flush (w);
  vt_setattr (w, w->curattr & ~ATTR_UL);
}

static inline void
vt_normal (vt_widget_t *w)
{
  vt_flush (w);
  vt_setattr (w, 0);
}

static inline void
vt_autowrap_on (vt_widget_t *w)
{
  w->autowrap = 1;
}

static inline void
vt_autowrap_off (vt_widget_t *w)
{
  w->autowrap = 0;
}

static inline void
vt_backspace (vt_widget_t *w)
{
  vt_flush (w);
  if (--w->curx < 0)
    w->curx = 0;
}

static inline void
vt_tab (vt_widget_t *w)
{
  vt_flush (w);
  w->curx = ((w->curx >> 3) + 1) << 3;
  if (w->curx >= w->wd)
    w->curx = w->wd-1;
}

static inline void
vt_lf (vt_widget_t *w)
{
  vt_flush (w);
  if (++w->cury >= w->ht) {
    vt_scroll (w);
    w->cury = w->ht-1;
  }
}

static inline void
vt_cr (vt_widget_t *w)
{
  vt_flush (w);
  w->curx = 0;
}

static inline void
vt_print (vt_widget_t *w, u_char c)
{
  vt_addchar (w, c);
  if (++w->curx >= w->wd) {
    vt_flush (w);
    if (!w->autowrap) {
      w->curx = w->wd-1;
    } else {
      w->curx = 0;
      if (++w->cury >= w->ht) {
	vt_scroll (w);
	w->cury = w->ht-1;
      }
    }
  }
}

/***************************** VT52 emulator *******************************/

static void
putc_esc_c (vt_widget_t *w, u_char c)
{
  w->outfn = putc_normal;
  vt_set_bgcol (w, c);
}

static void
putc_esc_b (vt_widget_t *w, u_char c)
{
  w->outfn = putc_normal;
  vt_set_fgcol (w, c);
}

static void
putc_esc_y2 (vt_widget_t *w, u_char c)
{
  w->outfn = putc_normal;
  vt_cursor_xpos (w, c);
}

static void
putc_esc_y1 (vt_widget_t *w, u_char c)
{
  w->outfn = putc_esc_y2;
  vt_cursor_ypos (w, c);
}

static void
putc_esc (vt_widget_t *w, u_char c)
{
  w->outfn = putc_normal;
  switch (c) {
    case 'A':	/* cursor up */
      vt_cursor_up (w);
      break;
    case 'B':	/* cursor down */
      vt_cursor_down (w);
      break;
    case 'C':	/* cursor right */
      vt_cursor_right (w);
      break;
    case 'D':	/* cursor left */
      vt_cursor_left (w);
      break;
    case 'E':	/* clear+home */
      vt_clear_screen (w);
      vt_cursor_home (w);
      break;
    case 'H':	/* home */
      vt_cursor_home (w);
      break;
    case 'I':	/* reverse index */
      vt_rev_index (w);
      break;
    case 'J':	/* erase to end of page */
      vt_clear_eop (w);
      break;
    case 'K':	/* erase to end of line */
      vt_clear_eol (w);
      break;
    case 'L':	/* insert line */
      vt_insert_line (w);
      break;
    case 'M':	/* delete line */
      vt_delete_line (w);
      break;
    case 'Y':	/* goto xy */
      w->outfn = putc_esc_y1;
      break;
    case 'b':	/* set fg col */
      w->outfn = putc_esc_b;
      break;
    case 'c':	/* set bg col */
      w->outfn = putc_esc_c;
      break;
    case 'd':	/* erase to start of page */
      vt_clear_sop (w);
      break;
    case 'e':	/* cursor on */
      vt_cursor_on (w);
      break;
    case 'f':	/* cursor off */
      vt_cursor_off (w);
      break;
    case 'j':	/* save cursor */
      vt_cursor_save (w);
      break;
    case 'k':	/* restore cursor */
      vt_cursor_restore (w);
      break;
    case 'o':	/* erase to start of line */
      vt_clear_sol (w);
      break;
    case 'p':	/* reverse on */
      vt_reverse_on (w);
      break;
    case 'q':	/* reverse off */
      vt_reverse_off (w);
      break;
    case 'v':	/* autowrap on */
      vt_autowrap_on (w);
      break;
    case 'w':	/*  autowrap off */
      vt_autowrap_off (w);
      break;

    case 'g':	/* bold on */
      vt_bold_on (w);
      break;
    case 'h':	/* normal */
      vt_normal (w);
      break;
    case 'i':	/* underline on */
      vt_underline_on (w);
      break;
    case 'm':	/* underline off */
      vt_underline_off (w);
      break;
    case 'n':	/* bold off */
      vt_bold_off (w);
      break;
  }
}

static void
putc_normal (vt_widget_t *w, u_char c)
{
  switch (c) {
    case 0:
      break;
    case 7: 	/* bell */
      vt_bell (w);
      break;
    case 8: 	/* backspace */
      vt_backspace (w);
      break;
    case 9: 	/* tab */
      vt_tab (w);
      break;
    case 10: 	/* linefeed */
      vt_lf (w);
      vt_cr (w);
      break;
    case 13:	/* carriage return */
      vt_cr (w);
      break;
    case 27:	/* escape */
      w->outfn = putc_esc;
      break;
    case 127:	/* delete */
      break;
    default:
      vt_print (w, c);
      break;
  }
}

/**************************** widget code **************************/

static long vt_query_geometry (widget_t *, long *, long *, long *, long *);

static long
vt_init (void)
{
  return 0;
}

static widget_t *
vt_create (widget_class_t *cp)
{
  vt_widget_t *wp = malloc (sizeof (vt_widget_t));
  if (!wp)
    return NULL;
  memset (wp, 0, sizeof (vt_widget_t));
  wp->w.class = wt_vt_class;
  wp->seltimer = -1;
  wp->clicktime = time (NULL);

  if (vt_loadfonts (wp, NULL, 0)) {
    free(wp);
    return NULL;
  }
  return (widget_t *)wp;
}

static long
vt_delete (widget_t *_w)
{
  vt_widget_t *w = (vt_widget_t *)_w;

  vt_exit (w);
  if (w->is_realized)
    w_delete (w->w.win);
  free (w);
  return 0;
}

static long
vt_close (widget_t *_w)
{
  vt_widget_t *w = (vt_widget_t *)_w;

  if (w->is_realized && w->is_open) {
    w_close (w->w.win);
    w->is_open = 0;
  }
  return 0;
}

static long
vt_open (widget_t *_w)
{
  vt_widget_t *w = (vt_widget_t *)_w;

  if (w->is_realized && !w->is_open) {
    w_open (w->w.win, w->w.x, w->w.y);
    w->is_open = 1;
  }
  return 0;
}

static long
vt_addchild (widget_t *parent, widget_t *w)
{
  return -1;
}

static long
vt_delchild (widget_t *parent, widget_t *w)
{
  return -1;
}

static long
vt_realize (widget_t *_w, WWIN *parent)
{
  vt_widget_t *w = (vt_widget_t *)_w;
  long x, y, wd, ht;

  if (w->is_realized)
    return -1;

  vt_query_geometry (_w, &x, &y, &wd, &ht);
  w->w.w = wd;
  w->w.h = ht;

  x = wd / w->fontwd;
  y = ht / w->fontht;
  if (vt_initialize (w, x, y, y + w->histsize))
    return -1;

  w->w.win = wt_create_window (parent, wd, ht,
	  W_NOBORDER|W_MOVE|EV_MOUSE, _w);
  if (!w->w.win)
    return -1;
  w->w.win->user_val = (long)w;

  w->is_open = 1;
  w->is_realized = 1;
  w_open (w->w.win, w->w.x, w->w.y);
  printf("vt_realize: w->fonts[0]: %s size=%d\n",w->fonts[0]->family,
				w->fonts[0]->height);
  w_setfont (w->w.win, w->fonts[0]);
  vt_cursor_show (w);
  return 0;
}

static long
vt_query_geometry (widget_t *_w, long *xp, long *yp, long *wdp, long *htp)
{
  vt_widget_t *w = (vt_widget_t *)_w;

  *xp = w->w.x;
  *yp = w->w.y;
  if (w->w.w > 0 && w->w.h > 0) {
    *wdp = w->w.w;
    *htp = w->w.h;
  } else {
    (*_w->class->query_minsize) (_w, wdp, htp);
  }
  w->w.w = *wdp;
  w->w.h = *htp;
  return 0;
}

static long
vt_query_minsize (widget_t *_w, long *wdp, long *htp)
{
  vt_widget_t *w = (vt_widget_t *)_w;

  if (!w->is_realized) {
    /* first time */
    *wdp = (w->usrcols > 0 ? w->usrcols : 80) * w->fontwd;
    *htp = (w->usrrows > 0 ? w->usrrows : 24) * w->fontht;
  } else {
    *wdp = 10 * w->fontwd;
    *htp = 10 * w->fontht;
  }
  *wdp = MAX (*wdp, w->usrwd);
  *htp = MAX (*htp, w->usrht);
  return 0;
}

static long
vt_reshape (widget_t *_w, long x, long y, long wd, long ht)
{
  vt_widget_t *w = (vt_widget_t *)_w;
  int ret = 0;

  if (x != w->w.x || y != w->w.y) {
    if (w->is_realized) {
      w_move (w->w.win, x, y);
    }
    w->w.x = x;
    w->w.y = y;
    ret = 1;
  }
  if (wd != w->w.w || ht != w->w.h) {
    wd /= w->fontwd;
    ht /= w->fontht;
    if (w->is_realized) {
      vt_resize (w, wd, ht, ht + w->histsize);
      w->w.w = w->wd * w->fontwd;
      w->w.h = w->ht * w->fontht;
    } else {
      w->w.w = wd * w->fontwd;
      w->w.h = ht * w->fontht;
    }
    ret = 1;
  }
  return ret;
}

static long
vt_setopt (widget_t *_w, long key, void *val)
{
  vt_widget_t *w = (vt_widget_t *)_w;
  short mask = 0;

  switch (key) {
    case WT_XPOS:
      if (vt_reshape (_w, *(long *)val, w->w.y, w->w.w, w->w.h))
        mask |= WT_CHANGED_POS;
      break;

    case WT_YPOS:
      if (vt_reshape (_w, w->w.x, *(long *)val, w->w.w, w->w.h))
        mask |= WT_CHANGED_POS;
      break;

    case WT_WIDTH:
      w->usrwd = MAX (0, *(long *)val);
      if (vt_reshape (_w, w->w.x, w->w.y, w->usrwd, w->w.h))
        mask |= WT_CHANGED_SIZE;
      break;

    case WT_HEIGHT:
      w->usrht = MAX (0, *(long *)val);
      if (vt_reshape (_w, w->w.x, w->w.y, w->w.w, w->usrht))
        mask |= WT_CHANGED_SIZE;
      break;

    case WT_VT_WIDTH:
      w->usrcols = *(long *)val;
      if (vt_reshape (_w, w->w.x, w->w.y, w->usrcols * w->fontwd, w->w.h))
	mask |= WT_CHANGED_SIZE;
      break;
      
    case WT_VT_HEIGHT:
      w->usrrows = *(long *)val;
      if (vt_reshape (_w, w->w.x, w->w.y, w->w.w, w->usrrows * w->fontwd))
	mask |= WT_CHANGED_SIZE;
      break;
      
    case WT_VT_HISTSIZE:
      if (!w->is_realized)
        w->histsize = *(long *)val;
      break;

    case WT_FONTSIZE:
      /* can't yet change font after realization :-( */
      if (w->is_realized) {
	printf("too late!\n");
        return -1;
      }
      vt_unloadfonts (w);
      printf("vt: WT_FONTSIZE %s size=%li\n",w->fontfamily,*(long *)val);
      vt_loadfonts (w, w->fontfamily, (short)(*(long *)val & 0xff));
      break;

    case WT_FONT:
      /* can't yet change font after realization :-( */
      if (w->is_realized) {
	printf("too late!\n");
        return -1;
      }
      vt_unloadfonts (w);
      printf("vt: WT_FONT %s size=%d\n",val,w->fontht);
      vt_loadfonts (w, (char *)val, w->fontht);
      break;

    case WT_VT_STRING:
      if (!w->is_realized)
        return -1;
      vt_output (w, ((wt_opaque_t *)val)->cp, ((wt_opaque_t *)val)->len);
      break;

    case WT_VT_HISTPOS:
      if (!w->is_realized)
        return -1;
      vt_vmove (w, *(long *)val);
      break;

    case WT_VT_VISBELL:
      w->visbell = *(long *)val;
      break;

    case WT_ACTION_CB:
      w->hist_cb = (void *)val;
      break;

    default:
      return -1;
  }
  if (mask && w->is_realized)
    wt_change_notify (_w, mask);
  return 0;
}

static long
vt_getopt (widget_t *_w, long key, void *val)
{
  vt_widget_t *w = (vt_widget_t *)_w;

  switch (key) {
    case WT_XPOS:
      *(long *)val = w->w.x;
      break;
      
    case WT_YPOS:
      *(long *)val = w->w.y;
      break;
      
    case WT_WIDTH:
      *(long *)val = w->w.w;
      break;
      
    case WT_HEIGHT:
      *(long *)val = w->w.h;
      break;

    case WT_VT_WIDTH:
      *(long *)val = w->wd ? w->wd : w->usrcols;
      break;

    case WT_VT_HEIGHT:
      *(long *)val = w->ht ? w->ht : w->usrrows;
      break;

    default:
      return -1;
  }
  return 0;
}

static void
select_timer_cb (long arg)
{
  vt_widget_t *w = (vt_widget_t *)arg;
  short x, y;

  w_querymousepos (w->w.win, &x, &y);
  vt_select_drag (w, x, y);
  w->seltimer = wt_addtimeout (100, select_timer_cb, arg);
}

static WEVENT *
vt_event (widget_t *_w, WEVENT *ev)
{
  vt_widget_t *w = (vt_widget_t *)_w;

  switch(ev->type) {
    case EVENT_MPRESS:
      if (ev->key & BUTTON_LEFT) {
	long tm = time (NULL);
	if (tm != w->clicktime) {
	  w->clicktime = tm;
	  w->nclicks = 0;
	}
	switch (++w->nclicks) {
	  case 1:
	    vt_select_start (w, ev->x, ev->y);
	    select_timer_cb ((long)w);
	    break;

	  case 2:
	    vt_select_word (w, ev->x, ev->y);
	    break;

	  case 3:
	    vt_select_line (w, ev->x, ev->y);
	    w->nclicks = 0;
	    break;
	}
      } else if (ev->key & BUTTON_RIGHT) {
	wt_opaque_t pasted;
	w_selection_t *sel = w_getselection (W_SEL_TEXT);
	if (sel) {
	  pasted.cp  = sel->data;
	  pasted.len = sel->len;
	  if (w->hist_cb)
	    (*w->hist_cb) (_w, -1, -1, -1, -1, &pasted);
	  w_freeselection (sel);
	} else {
	  w_beep ();
	}
      }
      break;

    case EVENT_MRELEASE:
      if (ev->key & BUTTON_LEFT) {
	if (w->nclicks == 1) {
	  if (ev->x >= 0 && ev->y >= 0)
	    vt_select_drag (w, ev->x, ev->y);
	  vt_select_end (w);
	}
      }
      break;

    default:
      return ev;
  }

  return NULL;
}

static long
vt_changes (widget_t *w, widget_t *w2, short changes)
{
  return 0;
}

static widget_class_t _wt_vt_class = {
  "vt", 0,
  vt_init,
  vt_create,
  vt_delete,
  vt_close,
  vt_open,
  vt_addchild,
  vt_delchild,
  vt_realize,
  vt_query_geometry,
  vt_query_minsize,
  vt_reshape,
  vt_setopt,
  vt_getopt,
  vt_event,
  vt_changes,
  vt_changes
};

widget_class_t *wt_vt_class = &_wt_vt_class;
