/*
 * wvt, terminal program for W based on wterm by TeSche.
 *
 * (W) '96, Kay Roemer, Torsten Scherer.
 *
 * $Id: wvt.c,v 1.3 1996/12/07 19:33:12 yak Exp $
 */

#include <stdio.h>
#include <assert.h>
#include <signal.h>
#include <sys/types.h>
#include <utmp.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <pwd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <Wlib.h>
#include <Wt.h>

#include "termtype.h"


#define	SMALLBUF 63
#define	LARGEBUF 1024


/*
 * some global variables
 */

static int pid, console, cfh, dh;
static int pipeh;
static char pty[SMALLBUF];
static int cblink = 0, visualbell = 0, debug = 0;
static struct winsize winsz;
static int winchflag = 0;

static long rows = 24, cols = 80, histsize = 100, curhistsize = 0;
static widget_t *top, *shell, *vt, *sb;
static long inphandler = 0;

extern void _write_utmp (const char *, const char *, const char *, u_long);
extern int  forkpty (int *, char *, struct termios *, struct winsize *);


/********************************************************************/

void sigpipe(int sig)
{
  /* this one musn't close the window */
  _write_utmp(pty, "", "", 0);
  kill(-pid, SIGHUP);
  _exit(sig);
}


void sigchld(int sig)
{
  _write_utmp(pty, "", "", 0);
  _exit(sig);
}


void sigquit(int sig)
{
  signal(sig, SIG_IGN);
  kill(-pid, sig);
}

void *mysignal(int signum, void *handler)
{
#ifdef __MINT__
  return signal(signum, handler);
#else
  struct sigaction sa, so;

  sa.sa_handler = handler;
#if defined(sun) || defined(__NetBSD__) || defined(__FreeBSD__)
  sa.sa_mask = 0;
  sa.sa_flags = 0;
#else
  sa.sa_mask = SA_NOMASK;
  sa.sa_flags = SA_RESTART;
#endif
  sigaction(signum, &sa, &so);

  return so.sa_handler;
#endif
}

/***********************************************************************/

static void
window_resized (void)
{
  struct winsize winsz;
  pid_t pgrp;

  winsz.ws_col = cols;
  winsz.ws_row = rows;

  if (!ioctl (pipeh, TIOCGPGRP, &pgrp) &&
      !ioctl (pipeh, TIOCSWINSZ, &winsz))
    kill (-pgrp, SIGWINCH);
}

static void
send_chars (char *cp, long len)
{
  long r;
  if (cp && len > 0) {
    r = write (pipeh, cp, len);
  }
}

static void
input_cb (long arg, fd_set *r, fd_set *w, fd_set *e)
{
  uchar buf[LARGEBUF];
  long in;

  /* is there tty data to be printed? */
  if (FD_ISSET (pipeh, r)) {
    if ((in = read (pipeh, buf, LARGEBUF)) > 0) {
      wt_opaque_t str;
      str.cp = buf;
      str.len = in;
      wt_setopt (vt, WT_VT_STRING, &str, WT_EOL);

      if (debug) {
	char *ptr = buf;

	while (in-- > 0) {
	  printf("wterm: received 0x%02x\n", *ptr++);
	}
      }
    }
  }

#ifdef __MINT__
  /* in any case: watch out for console output */
  if (console && FD_ISSET (cfh, r)) {
    if ((in = read (cfh, buf, LARGEBUF)) > 0) {
      wt_opaque_t str;
      str.cp = buf;
      str.len = in;
      wt_setopt (vt, WT_VT_STRING, &str, WT_EOL);
    }
  }
#endif

  if (FD_ISSET (pipeh, w)) {
    send_chars (NULL, 0);
  }
}

static void
key_cb (widget_t *w, WEVENT *ev)
{
  char c;

  if (ev->type == EVENT_KEY) {
    c = ev->key & 0x7f;
    send_chars (&c, 1);
  }
}

static void
vt_cb (widget_t *w, long pos, long size, long ncols, long nrows,
       wt_opaque_t *pasted)
{
  if (ncols > 0) {
    assert (size >= 0 && nrows > 0);
    rows = nrows;
    cols = ncols;
    window_resized ();
  }
  if (size >= 0) {
    long sbpos;
    wt_getopt (sb, WT_POSITION, &sbpos, WT_EOL);

    sbpos += size - curhistsize;
    curhistsize = size;
    size += rows;
    wt_setopt (sb,
	       WT_TOTAL_SIZE, &size,
	       WT_SIZE, &rows,
	       WT_PAGE_INC, &rows,
	       WT_POSITION, &sbpos, WT_EOL);
  }
  if (pos >= 0) {
    long sbpos;
    sbpos = curhistsize - pos;
    wt_setopt (sb, WT_POSITION, &sbpos, WT_EOL);
  }
  if (pasted) {
    send_chars (pasted->cp, pasted->len);
  }
}

static void
sb_cb (widget_t *w, long sbpos)
{
  long pos = curhistsize - sbpos;
  wt_setopt (vt, WT_VT_HISTPOS, &pos, WT_EOL);
}

static void
exit_cb (widget_t *w)
{
  sigquit (SIGHUP);
  wt_break (-1);
}

static int
init (char *fontname, long fsize, char *title, int xpos, int ypos)
{
  widget_t *hpane;
  fd_set rfds;
  long a, b;

  top = wt_init ();
  shell = wt_create (wt_shell_class, top);
  hpane = wt_create (wt_packer_class, shell);
  vt = wt_create (wt_vt_class, hpane);
  sb = wt_create (wt_scrollbar_class, hpane);

  a = (xpos == UNDEF ? UNDEF : (xpos < 0 ? 0 : xpos));
  b = (ypos == UNDEF ? UNDEF : (ypos < 0 ? 0 : ypos));
  wt_setopt (shell,
  	     WT_LABEL, title,
	     WT_XPOS, &a,
	     WT_YPOS, &b,
	     WT_ACTION_CB, exit_cb,
	     WT_EOL);

  a = 0;
  b = 1;
  wt_setopt (sb,
	     WT_TOTAL_SIZE, &rows,
             WT_SIZE, &rows,
	     WT_POSITION, &a,
             WT_LINE_INC, &b,
	     WT_PAGE_INC, &rows,
	     WT_ACTION_CB, sb_cb,
	     WT_EOL);

  if(wt_setopt (vt,
	     WT_FONTSIZE,&fsize,
  	     WT_FONT, fontname,
             WT_VT_WIDTH, &cols,
	     WT_VT_HEIGHT, &rows,
	     WT_VT_HISTSIZE, &histsize,
	     WT_VT_VISBELL, &visualbell,
	     WT_ACTION_CB, vt_cb,
	     WT_EOL)) {
	     perror("fontname");
	     exit(1);
  }

  wt_setopt (hpane,
  	     WT_PACK_WIDGET, vt,
  	     WT_PACK_INFO,   "-side left -expand 1 -fill both -padx 1",
  	     WT_PACK_WIDGET, sb,
  	     WT_PACK_INFO,   "-side left -fill both -padx 1",
  	     WT_EOL);

  wt_bind (NULL, EV_KEYS|EV_DEFAULT, key_cb);

  FD_ZERO (&rfds);
  FD_SET (pipeh, &rfds);
  if (console)
    FD_SET (cfh, &rfds);

  inphandler = wt_addinput (&rfds, NULL, NULL, input_cb, 0);
  wt_realize (top);

  if (xpos < 0 && xpos != UNDEF) {
    wt_getopt (shell, WT_WIDTH, &a, WT_EOL);
    xpos += wt_global.screen_wd - a - 5;
    wt_setopt (shell, WT_XPOS, &xpos, WT_EOL);
  }
  if (ypos < 0 && ypos != UNDEF) {
    wt_getopt (shell, WT_HEIGHT, &a, WT_EOL);
    ypos += wt_global.screen_ht - a - 19;
    wt_setopt (shell, WT_YPOS, &ypos, WT_EOL);
  }
  return 0;
}

/**********************************************************************/

static volatile void
usage (char *s)
{
  if (s) {
    fprintf(stderr, "error: %s\n", s);
  }
  puts("usage: wterm [-b] [-f <fontname>] [-s fontsize ] [-g <geometry>] "
       "[-h <histsize>] [-H] [-v] [-c] [program {args}]");
  exit(0);
}

int
main (long argc, char **argv)
{
  short xp, yp;
  char *fontname, *shell = NULL, *cptr, *geometry = NULL;
  struct passwd *pw;
  char buf[80], *ptr;
  short uid;
  char thesh[128];
  long fsize=11;
  int c;
#ifdef __MINT__
  int fd;
#endif
  extern char *optarg;
  extern int optind;

  /* just in case we're started in the background */
  signal(SIGTTOU, SIG_IGN);

  /* who am I? */
  if (!(pw = getpwuid((uid = getuid())))) {
    fprintf(stderr, "error: wterm can't determine determine your login name\n");
    exit(-1);
  }

  /*
   * which font shall I use?
   */

  fontname = "lucidat";

  /*
   * scan arguments...
   */

  console = 0;
  while ((c = getopt (argc, argv, "bcdf:s:g:h:vH")) != EOF) {
    switch (c) {
      case 'b':
	cblink = 1;
	break;

      case 'c':
	console = 1;
	break;

      case 'd':
	debug = 1;
	break;

      case 'f':
	fontname = optarg;
	ptr = strchr (fontname, '.');
	if (ptr)
	  *ptr = 0;
	break;

      case 's':
	fsize = atol(optarg);
	break;

      case 'g':
	geometry = optarg;
	break;

      case 'h':
	histsize = atol (optarg);
	if (histsize < 0)
	  histsize = 0;
	break;

      case 'H':
	/* this will never return */
	usage("");

      case 'v':
	visualbell = 1;
	break;

      case '?':
	usage("unknown option");
    }
  }
  argv += optind;

  /*
   * now *argv either points to a program to start or is zero
   */

  if (*argv) {
    shell = *argv;
  }
  if (!shell) {
    shell = getenv("SHELL=");
  }
  if (!shell) {
    shell = pw->pw_shell;
  }
  if (!shell) {
    shell = "/bin/sh";
  }

  if (!*argv) {
    /*
     * the '-' makes the shell think it is a login shell,
     * we leave argv[0] alone if it isn`t a shell (ie.
     * the user specified the program to run as an argument
     * to wterm.
     */
    cptr = strrchr(shell, '/');
    sprintf (thesh, "-%s", cptr ? cptr + 1 : shell);
    *--argv = thesh;
  }

  xp = UNDEF;
  yp = UNDEF;
  if (geometry) {
    short col, lin;
    scan_geometry(geometry, &col, &lin, &xp, &yp);
    cols = col < 1 ? 1 : col;
    rows = lin < 1 ? 1 : lin;
  }

  /*
   * what kind of terminal do we want to emulate?
   */

  putenv("TERM=wterm");

  /*
   * this one should enable us to get rid of an /etc/termcap entry for
   * both curses and ncurses, hopefully...
   */

  if (termcap_string) {
    sprintf (termcap_string + strlen (termcap_string), "li#%d:co#%d:",
	     rows, cols);
    putenv (termcap_string);
  }
  if (terminfo_string) {
    sprintf (terminfo_string + strlen (terminfo_string), "li#%d:co#%d:",
	     rows, cols);
    putenv (terminfo_string);
  }

  /*
   * create a pty
   */

  winsz.ws_col = cols;
  winsz.ws_row = rows;
  if ((pid = forkpty(&pipeh, pty, NULL, &winsz)) < 0) {
    printf("wterm: can't create pty\r\n");
    sleep(2);
    exit(1);
  }

  if ((ptr = rindex(pty, '/'))) {
    strcpy(pty, ptr + 1);
  }
  
  if (!pid) {
    int i;
    for (i = getdtablesize(); --i >= 3; )
      close (i);
    /*
     * SIG_IGN are not reset on exec()
     */
    for (i = NSIG; --i >= 0; )
      signal (i, SIG_DFL);
 
    /* caution: start shell with correct user id! */
    seteuid(getuid());
    setegid(getgid());

    /* this shall not return */
    execvp(shell, argv);

    /* oops? */
    printf("wterm: can't start shell\r\n");
    sleep(3);
    _exit(1);
  }

#ifdef __MINT__
  /*
   * the slave line opened in openpty() may still be our ctty.
   * so loose it.
   */
  fd = open ("/dev/null", O_RDWR);
  dup2 (fd, -1);
  close (fd);
#endif

  _write_utmp(pty, pw->pw_name, "", time(0));

  /* catch some signals */
  mysignal(SIGTERM, sigquit);
  mysignal(SIGHUP, sigquit);
  mysignal(SIGINT, sigquit);
  mysignal(SIGQUIT, sigquit);
  mysignal(SIGPIPE, sigpipe);
  mysignal(SIGCHLD, sigchld);

  /* prepare to catch console output */
  if (console) {

    /* for any OS chr$(7) might cause endless loops if catched from console */
    visualbell = 1;

#ifdef __MINT__
    if ((cfh = open("/dev/xconout2", O_RDONLY)) < 0) {
      printf("wterm: can't catch console output, device busy or not present\r\n");
      console = 0;
    }
#elif defined(sun) || defined(linux)
    console = 0;                     /* data will come to normal pipe handle */
    ioctl(pipeh, TIOCCONS, 0);
#else
    printf("wterm: can't catch console output on this operating system\r\n");
    console = 0;
#endif
  }

  /* caution: start shell with correct user id! */
  seteuid(getuid());
  setegid(getgid());


  fcntl (pipeh, F_SETFL, fcntl (pipeh, F_GETFL, 0) | O_NDELAY);
  if (console)
    fcntl (cfh, F_SETFL, fcntl (cfh, F_GETFL, 0) | O_NDELAY);
    
  sprintf(buf, "wvt: %s", shell);

  if (init (fontname, fsize, buf, xp, yp))
    _exit (1);
  
  if (debug)
    w_trace(1);

  wt_run ();
  return 0;
}
